#pragma once
#include <atlstr.h>
#include <crtdbg.h>

void PutLog(LPCTSTR entry);
CString ToString(bool res);

class MultiReaderSingleWriter
{
	HANDLE m_hSemRead;  // counts active readers
	HANDLE m_hEvtNoRead;  // prevents read operation

public:
	MultiReaderSingleWriter()
	{
		m_hSemRead = CreateSemaphore(NULL, 0, 1000, NULL);
		m_hEvtNoRead = CreateEvent(NULL, TRUE, FALSE, NULL);
	}

	~MultiReaderSingleWriter()
	{
		CloseHandle(m_hSemRead);
		CloseHandle(m_hEvtNoRead);
	}

	BOOL EnterRead()
	{
		PutLog(CString(TEXT("+ ENTER + ")) + TEXT(__FUNCTION__));
		DWORD waitRes = WaitForSingleObject(m_hEvtNoRead, INFINITE); // block until write completion
		_ASSERT(WAIT_OBJECT_0 == waitRes);	
		LONG count = 0;
		BOOL res = ReleaseSemaphore(m_hSemRead, 1, &count); // increments semaphore counter
		CString msg;
		msg.Format(TEXT("Semaphore count = %d, res = %d"), count, res);
		PutLog(msg);
		_ASSERT(res);
		PutLog(CString(TEXT("- RETURN - ")) + CString(__FUNCTION__) + TEXT(", res = ") + ToString(res));
		return res;
	}

	BOOL LeaveRead()
	{
		PutLog(CString(TEXT("+ ENTER + ")) + TEXT(__FUNCTION__));
		DWORD waitRes = WaitForSingleObject(m_hSemRead, 0); // decrement semaphore counter
		BOOL res = (WAIT_OBJECT_0 == waitRes);  // here not sure about the exact retcode :)
		_ASSERT(res);
		if (!res)
		{
			DWORD err = GetLastError();
			CString msg;
			msg.Format(TEXT("* waitRes = %d, last error: %d (0x%08X)"), waitRes, err, err);
			PutLog(msg);
		}
		PutLog(CString(TEXT("- RETURN - ")) + CString(__FUNCTION__) + TEXT(", res = ") + ToString(res));
		return res;
	}

	BOOL EnterWrite()
	{
		PutLog(CString(TEXT("+ ENTER + ")) + TEXT(__FUNCTION__));
		BOOL res = ResetEvent(m_hEvtNoRead); // prevents new reads
		_ASSERT(res);

		DWORD waitRes = 0;
		do
		{
			// loop until last reader gone
			LONG count = 0;
			res = ReleaseSemaphore(m_hSemRead, 1, &count);
			_ASSERT(res);
			waitRes = WaitForSingleObject(m_hSemRead, 0);
			if (count)
				Sleep(10);
			else if (WAIT_OBJECT_0 == waitRes)
			{
				res = TRUE;
				break;
			}
			else
			{
				res = FALSE;
				break;
			}
		} while (TRUE);

		PutLog(CString(TEXT("- RETURN - ")) + CString(__FUNCTION__) + TEXT(", res = ") + ToString(res));
		return res;
	}

	BOOL LeaveWrite()
	{
		PutLog(CString(TEXT("+ ENTER + ")) + TEXT(__FUNCTION__));
		BOOL res = SetEvent(m_hEvtNoRead); // release waiting readers
		_ASSERT(res);
		PutLog(CString(TEXT("- RETURN - ")) + CString(__FUNCTION__) + TEXT(", res = ") + ToString(res));
		return res;
	}
};